import logging
import sys

import typer
from rich.console import Console

from .logging import setup_logging
from .python_cli import get_active_python_info
from .releases import load_releases

if sys.version_info < (3, 9):
    from typing_extensions import Annotated
else:
    from typing import Annotated

RICH_CONSOLE = Console()


app = typer.Typer(name="pirel")
logger = logging.getLogger("pirel")


def logging_callback(ctx: typer.Context, verbosity: int):
    if ctx.resilient_parsing:
        return

    setup_logging(verbosity)
    return verbosity


VERBOSE_OPTION = Annotated[
    int,
    typer.Option(
        "--verbose",
        "-v",
        help="Enable verbose logging; can be supplied multiple times to increase verbosity.",
        count=True,
        callback=logging_callback,
    ),
]


def print_releases():
    """Prints all Python releases as a table."""
    py_info = get_active_python_info()
    py_version = py_info.version if py_info else None

    releases = load_releases()
    RICH_CONSOLE.print(releases.to_table(py_version), new_line_start=True)


# This callback is a hack to "redirect" to the `list` command if no command is passed.
# This will be removed in a future version.
@app.callback(invoke_without_command=True)
def redirect_no_command_to_list(
    ctx: typer.Context,
    verbose: Annotated[int, typer.Option("--verbose", "-v", count=True)] = 0,
):
    if not ctx.invoked_subcommand:
        setup_logging(verbose)
        logger.warning(
            "Invoking `pirel` without a command is deprecated"
            " and will be removed in a future version."
        )
        logger.warning("Please use `pirel list` instead.")
        print_releases()


@app.command("list")
def list_releases(verbose: VERBOSE_OPTION = 0):
    """Lists all Python releases in a table. Your active Python interpreter is highlighted."""
    print_releases()


@app.command("check")
def check_release(verbose: VERBOSE_OPTION = 0):
    """Shows release information about your active Python interpreter.

    If no active Python interpreter is found, the program exits with code 2.
    """
    py_info = get_active_python_info()
    if not py_info:
        logger.error("Could not find active Python interpreter in PATH.")
        raise typer.Exit(code=2)

    releases = load_releases()
    active_release = releases[py_info.version.as_release]

    RICH_CONSOLE.print(f"\n{active_release}", highlight=False)


if __name__ == "__main__":
    app()  # pragma: no cover



from __future__ import annotations

import datetime
import json
import logging
import urllib.request
from typing import Any, Optional

import humanize
from rich.table import Table

from . import python_cli

DATE_NOW = datetime.date.today()
RELEASE_CYCLE_URL = "https://raw.githubusercontent.com/python/devguide/refs/heads/main/include/release-cycle.json"

STATUS_TO_TEXT = {
    "feature": "is [bold]not released yet[/bold] and still accepts new features",
    "prerelease": "is [bold]not released yet[/bold] and still accepts new fixes",
    "bugfix": "is actively maintained (bugfixes) and has security support for more than {eol_delta}",
    "security": "has security support for more than {eol_delta}",
    "end-of-life": "has reached end-of-file! Please upgrade to a newer version of Python",
}

STATUS_TO_EMOJI = {
    "feature": ":sparkles:",
    "prerelease": ":eyes:",
    "bugfix": ":rocket:",
    "security": ":heavy_check_mark:",
    "end-of-life": ":warning:",
}

logger = logging.getLogger("pirel")


def parse_date(date_str: str) -> datetime.date:
    if len(date_str) == len("yyyy-mm"):
        # We need a full yyyy-mm-dd, so let's approximate
        return datetime.date.fromisoformat(date_str + "-01")
    return datetime.date.fromisoformat(date_str)


def date_style(date: datetime.date) -> str:
    """Returns the style for a date for rich table."""
    if date > DATE_NOW:
        # Future, add italics
        return "italic"
    return ""


def status_style(status: str) -> str:
    if status == "end-of-life":
        return "red"
    elif status == "security":
        return "yellow"
    elif status == "bugfix":
        return "green"
    elif status == "prerelease":
        return "blue"
    elif status == "feature":
        return "magenta"
    else:
        raise ValueError(f"Unkown status {status}")


def eol_color(eol: datetime.date) -> str:
    if DATE_NOW >= eol:
        return "red"
    elif datetime.timedelta(days=30 * 6) + DATE_NOW > eol:
        return "dark_orange"
    elif datetime.timedelta(days=365) + DATE_NOW > eol:
        return "yellow"
    else:
        return "green"


def wrap_style(text: str, style: str) -> str:
    if not style:
        return text
    return f"[{style.strip()}]{text}[/{style.strip()}]"


class PythonRelease:
    def __init__(self, version: str, data: dict):
        self._version = version
        self._status = data["status"]
        self._released = parse_date(data["first_release"])
        self._end_of_life = parse_date(data["end_of_life"])

    def __str__(self) -> str:
        status_info = STATUS_TO_TEXT[self._status].format(
            eol_delta=humanize.naturaldelta(self._end_of_life - DATE_NOW)
        )
        main_style = status_style(self._status)
        _eol_color = eol_color(self._end_of_life)
        return (
            f"{STATUS_TO_EMOJI[self._status]} [{main_style} bold]"
            f"You are using Python {self.version}[/] which {status_info}"
            f" [{_eol_color}](EOL {self._end_of_life})[/{_eol_color}]"
        )

    @property
    def version(self) -> str:
        return self._version

    @property
    def version_tuple(self) -> tuple[int, int]:
        return self._version.split(".")

    @property
    def status(self) -> str:
        return wrap_style(self._status, status_style(self._status))

    @property
    def released(self) -> str:
        return wrap_style(str(self._released), style=date_style(self._released))

    @property
    def end_of_life(self) -> str:
        font_style = date_style(self._end_of_life)
        color = eol_color(self._end_of_life)
        return wrap_style(str(self._end_of_life), style=f"{font_style} {color}")


class PythonReleases:
    def __init__(self, releases_data: dict[str, dict[str, Any]]) -> None:
        self.releases = {
            version: PythonRelease(version, data)
            for version, data in releases_data.items()
        }

    def __getitem__(self, version: str) -> PythonRelease:
        return self.releases[version]

    def to_table(
        self, active_python_version: Optional[python_cli.PythonVersion] = None
    ) -> Table:
        table = Table(title="Python Releases")

        table.add_column("Version", justify="right", style="cyan", no_wrap=True)
        table.add_column("Status", justify="right", no_wrap=True)
        table.add_column("Released", justify="right", style="grey66", no_wrap=True)
        table.add_column("End-of-life", justify="right", no_wrap=True)

        for release in self.releases.values():
            row_style = None
            _version = release.version
            if (
                active_python_version
                and active_python_version.as_release == release.version
            ):
                _version = f"* [bold]{release.version}[/bold]"
                row_style = "bold"

            table.add_row(
                _version,
                release.status,
                release.released,
                release.end_of_life,
                style=row_style,
            )

        return table


def load_releases() -> PythonReleases:
    """Downloads the release cycle data (JSON) from the Python devguide repo."""
    logger.debug(f"Downloading Python release cycle data from URL {RELEASE_CYCLE_URL}")
    with urllib.request.urlopen(RELEASE_CYCLE_URL) as f:
        releases_data = json.load(f)

    return PythonReleases(releases_data)



import datetime
import logging
import pathlib
import re
from unittest import mock

import pytest
from rich.console import Console
from typer.testing import CliRunner

from pirel.cli import app
from pirel.python_cli import PythonVersion

runner = CliRunner()
RELEASES_TABLE = """
┏━━━━━━━━━┳━━━━━━━━━━━━━┳━━━━━━━━━━━━┳━━━━━━━━━━━━━┓
┃ Version ┃      Status ┃   Released ┃ End-of-life ┃
┡━━━━━━━━━╇━━━━━━━━━━━━━╇━━━━━━━━━━━━╇━━━━━━━━━━━━━┩
│    3.14 │     feature │ 2025-10-01 │  2030-10-01 │
│    3.13 │      bugfix │ 2024-10-07 │  2029-10-01 │
│    3.12 │      bugfix │ 2023-10-02 │  2028-10-01 │
│    3.11 │    security │ 2022-10-24 │  2027-10-01 │
│    3.10 │    security │ 2021-10-04 │  2026-10-01 │
│     3.9 │    security │ 2020-10-05 │  2025-10-01 │
│     3.8 │ end-of-life │ 2019-10-14 │  2024-10-07 │
│     3.7 │ end-of-life │ 2018-06-27 │  2023-06-27 │
│     3.6 │ end-of-life │ 2016-12-23 │  2021-12-23 │
│     3.5 │ end-of-life │ 2015-09-13 │  2020-09-30 │
│     3.4 │ end-of-life │ 2014-03-16 │  2019-03-18 │
│     3.3 │ end-of-life │ 2012-09-29 │  2017-09-29 │
│     3.2 │ end-of-life │ 2011-02-20 │  2016-02-20 │
│     2.7 │ end-of-life │ 2010-07-03 │  2020-01-01 │
│     3.1 │ end-of-life │ 2009-06-27 │  2012-04-09 │
│     3.0 │ end-of-life │ 2008-12-03 │  2009-06-27 │
│     2.6 │ end-of-life │ 2008-10-01 │  2013-10-29 │
└─────────┴─────────────┴────────────┴─────────────┘
""".strip()

PYVER_TO_CHECK_OUTPUT = {
    "3.8": ":warning: You are using Python 3.8 which has reached end-of-file! Please upgrade to a newer version of Python (EOL 2024-10-07)",
    "3.9": ":heavy_check_mark: You are using Python 3.9 which has security support for more than 10 months (EOL 2025-10-01)",
    "3.10": ":heavy_check_mark: You are using Python 3.10 which has security support for more than 1 year, 10 months (EOL 2026-10-01)",
    "3.11": ":heavy_check_mark: You are using Python 3.11 which has security support for more than 2 years (EOL 2027-10-01)",
    "3.12": ":rocket: You are using Python 3.12 which is actively maintained (bugfixes) and has security support for more than 3 years (EOL 2028-10-01)",
    "3.13": ":rocket: You are using Python 3.13 which is actively maintained (bugfixes) and has security support for more than 4 years (EOL 2029-10-01)",
}


@pytest.fixture(scope="module")
def mock_rich_no_wrap():
    with mock.patch("pirel.cli.RICH_CONSOLE", Console(soft_wrap=True, emoji=False)):
        yield


@pytest.fixture
def mock_release_cycle_file():
    date_freeze = "2024-11-03"
    data_path = (
        pathlib.Path(__file__).parent / "data" / f"release-cycle_{date_freeze}.json"
    )
    with open(data_path) as file:
        with mock.patch("pirel.releases.urllib.request.urlopen") as mock_urlopen:
            # Mock call to release cycle data
            mock_urlopen.return_value.__enter__.return_value = file

            # Mock date for reproducability
            with mock.patch(
                "pirel.releases.DATE_NOW", datetime.date.fromisoformat(date_freeze)
            ):
                yield


@pytest.fixture
def releases_table():
    pyver = PythonVersion.this()
    # Add asterisk to active Python version
    table = re.sub(rf"  {pyver.as_release}", f"* {pyver.as_release}", RELEASES_TABLE)
    return table


@pytest.mark.parametrize("args, log_count", [(None, 2), ("list", 0)])
def test_pirel_list(
    args, log_count, mock_rich_no_wrap, mock_release_cycle_file, releases_table, caplog
):
    caplog.set_level(logging.WARNING)

    # Call CLI
    result = runner.invoke(app, args)
    assert result.exit_code == 0, result.stdout

    # Check output
    output = result.stdout.strip()
    heading, *table = output.splitlines()
    table = "\n".join(table)

    assert heading.strip() == "Python Releases"
    assert table.strip() == releases_table

    logs = caplog.messages
    assert len(logs) == log_count
    # If called without args, check that the warning is emitted
    if args is None:
        assert logs[-1] == "Please use `pirel list` instead."


def test_pirel_check(mock_rich_no_wrap, mock_release_cycle_file):
    pyver = PythonVersion.this()

    # Call CLI
    result = runner.invoke(app, "check")
    assert result.exit_code == 0, result.stdout

    # Check output
    output = result.stdout.strip()
    assert output == PYVER_TO_CHECK_OUTPUT[pyver.as_release]


def test_pirel_check_no_interpreter():
    with mock.patch("pirel.cli.get_active_python_info") as m_py_info:
        m_py_info.return_value = None

        # Call CLI
        result = runner.invoke(app, "check")
        assert result.exit_code == 2, result.stdout



import logging

from rich.logging import RichHandler


def setup_logging(verbosity: int = 0):
    """Sets up the basic logging configuration.

    Args:
        verbosity (int): Configures the log level which defaults to WARNING.
            A `verbosity` of `0` maps to WARNING, `1` -> INFO, and `2` (or more)
            -> DEBUG. Defaults to `0`.
    """
    base_loglevel = logging.WARNING  # 30
    # Calculate the log level based on verbosity, i.e. subtract intervals of 10
    # from level WARNING (30).
    loglevel = max(base_loglevel - (verbosity * 10), logging.DEBUG)
    logging.basicConfig(
        level=loglevel,
        format="%(message)s",
        handlers=[RichHandler(show_path=False, show_time=False)],
    )



from __future__ import annotations

import logging
import re
import subprocess
import sys
from dataclasses import dataclass

PYTHON_VERSION_RE = re.compile(r"Python ([23])\.(\d+)\.(\d+)")
logger = logging.getLogger("pirel")


@dataclass(frozen=True)
class PythonVersion:
    major: int
    minor: int
    patch: int

    @classmethod
    def from_cli(cls, version: str) -> PythonVersion:
        match = PYTHON_VERSION_RE.match(version)
        if not match:
            raise ValueError(
                f"The Python version output {version!r} "
                f"does not match the regex {PYTHON_VERSION_RE.pattern!r}"
            )
        major, minor, patch = match.groups()
        return PythonVersion(major, minor, patch)

    @classmethod
    def this(cls) -> PythonVersion:
        return cls(*sys.version_info[:3])

    @property
    def as_release(self) -> str:
        return f"{self.major}.{self.minor}"

    def __str__(self) -> str:
        return f"{self.major}.{self.minor}.{self.patch}"

    def __repr__(self) -> str:
        return f"PythonVersion({self.major}.{self.minor}.{self.patch})"


@dataclass(frozen=True)
class ActivePythonInfo:
    cmd: str
    path: str
    version: PythonVersion


def get_active_python_info() -> ActivePythonInfo | None:
    version_out = None
    for py_exe in ("python", "python3", "python2"):
        try:
            logger.debug(f"Trying command `{py_exe}`")
            # TODO: Check if there is a better way to inspect the active Python version
            version_out = subprocess.run((py_exe, "--version"), capture_output=True)
            version = PythonVersion.from_cli(version_out.stdout.decode())
            break
        except (FileNotFoundError, ValueError):
            logger.debug(
                f"Failed to get Python version from command {py_exe!r}", exc_info=True
            )

    if version_out is None:
        logger.warning("Could not find an active Python interpreter")
        return None

    path_out = subprocess.run(
        (py_exe, "-c", "import sys; print(sys.executable)"), capture_output=True
    )
    path = path_out.stdout.decode().strip()
    logger.info(f"Found active Python interpreter at {path!r} (command {py_exe!r})")

    return ActivePythonInfo(cmd=py_exe, path=path, version=version)
